/*******************************************************************************
*    DOGM-Library 1.0.                                                         *
*    Copyright (C) 2010-averyfarwaydate Luca Bertoncello                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <util/delay.h>
#include "inc/dogm.h"

// Instructions
#define OPCODE_CLEAR_DISPLAY        0b00000001
#define OPCODE_FUNCTION_SET_INIT_0  0b00110011
#define OPCODE_FUNCTION_SET_INIT_1  0b00110010
#define OPCODE_FUNCTION_SET_INIT_2  0b00101001
#define OPCODE_INSTRUCTION_SET_0    0b00101000
#define OPCODE_INSTRUCTION_SET_1    0b00101001
#define OPCODE_BIAS_SET             0b00010101
#define OPCODE_POWER_CONTROL5       0b01010011
#define OPCODE_POWER_CONTROL33      0b01010110
#define OPCODE_FOLLOWER_CONTROL     0b01101100
#define OPCODE_CONTRAST_SET         0b01110000
#define OPCODE_DISPLAY_ON           0b00001100
#define OPCODE_ENTRY_MODE           0b00000110
#define OPCODE_GOTO                 0b10000000

#define lcdSetEnable()      LCD_E_PORT  |=  _BV(LCD_ENABLE);
#define lcdClearEnable()    LCD_E_PORT  &= ~_BV(LCD_ENABLE);
#define lcdSetRW()          LCD_RW_PORT |=  _BV(LCD_RW)
#define lcdClearRW()        LCD_RW_PORT &= ~_BV(LCD_RW)
#define lcdSetRegSelect()   LCD_RS_PORT |=  _BV(LCD_REGSELECT)
#define lcdClearRegSelect() LCD_RS_PORT &= ~_BV(LCD_REGSELECT)

/**
 * Reads a byte from LCD-controller
 *
 * @param uint8_t rs        0: read busy flag / address counter
 *                          1: read data
 * @return uint8_t          the data
 */
uint8_t lcdReadByte(uint8_t rs)
{
  uint8_t data;

  if(rs)
    lcdSetRegSelect();                      // RS=1: read data
  else
    lcdClearRegSelect();                    // RS=0: read busy flag
  lcdSetRW();                               // RW=1: read mode

  LCD_DATA_DDR &= ~(_BV(LCD_DATA7) | _BV(LCD_DATA6) | _BV(LCD_DATA5) | _BV(LCD_DATA4));

  lcdSetEnable();
  _delay_ms(2);
  data = 0;
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA4)) data |= 0x10;
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA5)) data |= 0x20;
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA6)) data |= 0x40;
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA7)) data |= 0x80;
  lcdClearEnable();

  _delay_ms(2);

  lcdSetEnable();
  _delay_ms(2);
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA4)) data |= 0x01;
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA5)) data |= 0x02;
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA6)) data |= 0x04;
  if(bit_is_set(LCD_DATA_PIN, LCD_DATA7)) data |= 0x08;
  lcdClearEnable();

  return data;
}

/**
 * Sends a byte to the LCD-controller
 *
 * @param char data         the byte
 */
void lcdSendByte(char data)
{
  char port_buffer = LCD_DATA_PORT;
  char nibble, byte;

  lcdClearRW();
  LCD_DATA_DDR |= _BV(LCD_DATA7) | _BV(LCD_DATA6) | _BV(LCD_DATA5) | _BV(LCD_DATA4);
  nibble = (data & 0xF0) >> 4;
  byte = ((nibble & 0x01) << LCD_DATA4) | (((nibble & 0x02) >> 1) << LCD_DATA5) | (((nibble & 0x04) >> 2) << LCD_DATA6) | (((nibble & 0x08) >> 3) << LCD_DATA7);

  LCD_DATA_PORT |= byte;
  lcdSetEnable();
  _delay_ms(2);

  LCD_DATA_PORT = port_buffer;
  _delay_ms(2);

  nibble = data & 0x0F;
  byte = ((nibble & 0x01) << LCD_DATA4) | (((nibble & 0x02) >> 1) << LCD_DATA5) | (((nibble & 0x04) >> 2) << LCD_DATA6) | (((nibble & 0x08) >> 3) << LCD_DATA7);

  LCD_DATA_PORT |= byte;
  lcdSetEnable();
  _delay_ms(2);

  LCD_DATA_PORT = port_buffer;
  _delay_ms(2);
}

/**
 * Sends an instruction to the LCD-controller
 *
 * @param char instruction  the instruction
 */
void lcdSendInstruction(char instruction)
{
  //RS low = instruction
  lcdClearEnable();
  lcdClearRW();
  lcdClearRegSelect();
  lcdSendByte(instruction);
}

/**
 * Writes a char on the LCD-display
 *
 * @param char c            the character
 */
void lcdPutc(char c)
{
  //RS high = data
  lcdClearEnable();
  lcdClearRW();
  lcdSetRegSelect();
  lcdSendByte(c);
}

/**
 * Writes a string on the LCD-display.
 * \n will accepted as Carriage Return
 *
 * @param char *str         the string
 */
void lcdPuts(char *str)
{
  uint8_t l, pos;

  for(l = 0; str[l] != 0; l++)
    if(str[l] == '\n')
    {
      pos = lcdReadByte(0);
      lcdGotoXY(0, (pos / LCD_COLS) + 1);
    }
    else
      lcdPutc(str[l]);
}

/**
 * Delete the content of the LCD-display
 */
void lcdClear()
{
  lcdSendInstruction(OPCODE_CLEAR_DISPLAY);
}

/**
 * Sets the contrast of the LCD-display
 *
 * @param uint8_t contrast  the contrast (4 bits!!)
 */
void lcdContrast(uint8_t contrast)
{
  lcdSendInstruction(OPCODE_FUNCTION_SET_INIT_2);
  lcdSendInstruction(OPCODE_CONTRAST_SET | (contrast & 0x0F));
}

/**
 * Moves the cursor to the position x/y
 *
 * @param uint8_t x         the horizontal position
 * @param uint8_t y         the vertical position
 */
void lcdGotoXY(uint8_t x, uint8_t y)
{
  uint8_t pos;

  if(x > LCD_COLS)
    x = 0;
  if(y > LCD_LINES)
    y = 0;
  pos = y * LCD_COLS + x;
  lcdSendInstruction(OPCODE_GOTO | pos);
}

/**
 * Reads the current cursor's position
 *
 * @param uint8_t *x        the horizontal position (as reference)
 * @param uint8_t *y        the vertical position (as reference)
 */
void lcdGetXY(uint8_t *x, uint8_t *y)
{
  uint8_t pos;

  pos = lcdReadByte(0);
  *y = (pos / LCD_COLS);
  *x = pos - (*y * LCD_COLS);
}

/**
 * Initialises the LCD-controller
 *
 * @param uint8_t power     POWER5V or POWER3V3
 * @param uint8_t contrast  the initial contrast
 * @param uint8_t cursor    1 if the cursor should be displayed, 0 otherwise
 * @param uint8_t blink     1 if the cursor should blink, 0 otherwise
 */
void lcdInit(uint8_t power, uint8_t contrast, uint8_t cursor, uint8_t blink)
{
  _delay_ms(50);

  LCD_BACKLIGHT_DDR |= _BV(LCD_BACKLIGHT);
  LCD_E_DDR |= _BV(LCD_ENABLE);
  LCD_RW_DDR |= _BV(LCD_RW);
  LCD_RS_DDR |= _BV(LCD_REGSELECT);
  LCD_DATA_DDR |= _BV(LCD_DATA7) | _BV(LCD_DATA6) | _BV(LCD_DATA5) | _BV(LCD_DATA4);

  //4 Bits mode
  lcdSendInstruction(OPCODE_FUNCTION_SET_INIT_0);
  lcdSendInstruction(OPCODE_FUNCTION_SET_INIT_1);
  lcdSendInstruction(OPCODE_FUNCTION_SET_INIT_2);

  lcdSendInstruction(OPCODE_BIAS_SET);
  if(power == POWER5V)
    lcdSendInstruction(OPCODE_POWER_CONTROL5);
  else
    lcdSendInstruction(OPCODE_POWER_CONTROL33);

  lcdSendInstruction(OPCODE_FOLLOWER_CONTROL);
  lcdSendInstruction(OPCODE_CONTRAST_SET | (contrast & 0x0F));
  lcdSendInstruction(OPCODE_INSTRUCTION_SET_0);
  lcdSendInstruction(OPCODE_DISPLAY_ON | (cursor & 0x01) << 1 | (blink & 0x01));
  lcdClear();
  lcdSendInstruction(OPCODE_ENTRY_MODE);
}
